"""
Salt execution module for VMC Security Rules
Provides methods to Create, Read, Update and Delete security rules.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_security_rules"


def __virtual__():
    return __virtualname__


def _create_payload_for_security_rule(rule_id, domain_id, user_input):
    """
    This function creates the payload based on the template and user input passed
    """
    template_data = getattr(
        vmc_templates, "create_security_rules_" + domain_id, vmc_templates.create_security_rules_cgw
    )
    data = vmc_request.create_payload_for_request(template_data, user_input)
    data["id"] = data["display_name"] = rule_id
    return data


def get(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    verify_ssl=True,
    cert=None,
    sort_by=None,
    sort_ascending=None,
    page_size=None,
    cursor=None,
):
    """
    Retrieves security rules for Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_rules.get hostname=nsxt-manager.local domain_id=mgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security rules should be retrieved

    domain_id
        The domain_id for which the security rules should be retrieved. Possible values: mgw, cgw

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order

    page_size
        (Optional) Maximum number of results to return in this page

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    """

    log.info("Retrieving %s Security rules for SDDC %s", domain_id, sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/gateway-policies/default/rules"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_rules.get",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_id(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves given security rule from the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_rules.get_by_id hostname=nsxt-manager.local domain_id=mgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security rule should be retrieved

    domain_id
        The domain_id for which the security rule should be retrieved. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be retrieved from SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving Security rule %s for SDDC %s", rule_id, sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/gateway-policies/default/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id, rule_id=rule_id
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_rules.get_by_id",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
):
    """
    Deletes given security rule from the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_rules.delete hostname=nsxt-manager.local domain_id=mgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the security rule should be deleted

    domain_id
        The domain_id for which the security rule should be deleted. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be deleted from SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Deleting Security rule %s for SDDC %s", rule_id, sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/gateway-policies/default/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id, rule_id=rule_id
    )

    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_rules.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    tag=None,
    logged=None,
    disabled=None,
    notes=None,
    sequence_number=None,
    tags=vmc_constants.VMC_NONE,
):
    """
    Creates security rule for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_rules.create hostname=nsxt-manager.local domain_id=mgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security rule should be added

    domain_id
        The domain_id for which the security rule should be added. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be added to given SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    Note: Both source_groups and destination_groups can not be ["ANY"] when domain_id=mgw

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for domain_id=cgw are: ALLOW, DROP, REJECT
        Possible Values for domain_id=mgw are: ALLOW

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    disabled
        (Optional) Flag to disable the rule. Default is enabled.

    notes
        (Optional) Text for additional notes on changes.

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    Example values:

        .. code::

            {
                "sequence_number": 0,
                "source_groups": [
                    "ANY"
                ],
                "services": ["/infra/services/HTTPS"],
                "logged": false,
                "disabled": false,
                "destination_groups": [
                    "/infra/domains/mgw/groups/VCENTER"
                ],
                "scope": [
                    "/infra/tier-1s/mgw"
                ],
                "action": "ALLOW",
                "tag": "",
                "notes": "",
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ]
            }

    """

    log.info("Creating Security rule %s for SDDC %s", rule_id, sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/gateway-policies/default/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id, rule_id=rule_id
    )

    allowed_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "tag": tag,
        "logged": logged,
        "disabled": disabled,
        "notes": notes,
        "sequence_number": sequence_number,
        "tags": tags,
    }

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = _create_payload_for_security_rule(rule_id, domain_id, req_data)
    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_rules.create",
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    tag=None,
    logged=None,
    disabled=None,
    notes=None,
    sequence_number=None,
    tags=vmc_constants.VMC_NONE,
    display_name=None,
):
    """
    Updates security rule for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_rules.update hostname=nsxt-manager.local domain_id=mgw ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security rule belongs to

    domain_id
        The domain_id for which the security rule belongs to. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be updated for given SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    Note: Both source_groups and destination_groups can not be ["ANY"] when domain_id=mgw

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for domain_id=cgw are: ALLOW, DROP, REJECT
        Possible Values for domain_id=mgw are: ALLOW

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    disabled
        (Optional) Flag to disable the rule. Default is enabled.

    notes
        (Optional) Text for additional notes on changes.

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    display_name
        Identifier to use when displaying entity in logs or GUI

    Example values:

        .. code::

            {
                "display_name": "vCenter Inbound Rule"
                "sequence_number": 0,
                "source_groups": [
                    "ANY"
                ],
                "services": ["/infra/services/HTTPS"],
                "logged": false,
                "disabled": false,
                "destination_groups": [
                    "/infra/domains/mgw/groups/VCENTER"
                ],
                "scope": [
                    "/infra/tier-1s/mgw"
                ],
                "action": "ALLOW",
                "tag": "",
                "notes": "",
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ]
            }

    """

    log.info("Updating Security rule %s for SDDC %s", rule_id, sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/gateway-policies/default/rules/{rule_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id, rule_id=rule_id
    )

    # fetch the security rule for the given rule_id
    existing_data = get_by_id(
        hostname,
        refresh_key,
        authorization_host,
        org_id,
        sddc_id,
        domain_id,
        rule_id,
        verify_ssl,
        cert,
    )

    if vmc_constants.ERROR in existing_data:
        return existing_data

    allowed_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "tag": tag,
        "logged": logged,
        "disabled": disabled,
        "notes": notes,
        "sequence_number": sequence_number,
        "display_name": display_name,
        "tags": tags,
    }

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = vmc_request.create_payload_for_request(
        vmc_templates.update_security_rules, req_data, existing_data
    )
    return vmc_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_rules.update",
        responsebody_applicable=False,
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )



# SPDX-License-Identifier: Apache-2.0
import json
import logging

import saltext.vmware.modules.storage_policies as policy_module
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.drift as drift

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "storage_policy"
__proxyenabled__ = ["storage_policy"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def config(name, config, service_instance=None, profile=None):
    """
    Get/Set storage policies configuration based on drift report.

    name
        Name of configuration. (required).

    config
        List of objects with configuration values. (required).

    .. code-block:: yaml

        storage_policies_drift:
            storage_policy.config:
                - profile: vcenter
                - config:
                - policyName: Performance - Thick
                    constraints:
                    - name: VSAN
                        replicaPreference: 1 failures - RAID-1 (Mirroring)
                        hostFailuresToTolerate: 2
                        checksumDisabled: true
                        stripeWidth: 2
                        proportionalCapacity: 50
                        cacheReservation: 0
                - policyName: New VM Storage Policy
                    constraints:
                    - name: VSAN
                        hostFailuresToTolerate: 2
                        replicaPreference: 2 failures - RAID-1 (Mirroring)
                        proportionalCapacity: 1

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """

    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # Clone config input to a Map.
    # Can be used to transform input to internal objects and do validation if needed
    new_config = {}
    for policy_config in config:
        # Create full representation of the object, default or empty values
        new_policy_config = {"constraints": {}}
        # Transform / Validate input vs object, e.g. constraints section
        if "constraints" in policy_config:
            for constraint in policy_config["constraints"]:
                subProfileName = constraint["name"]
                new_policy_config["constraints"][subProfileName] = {}
                for key in constraint.keys():
                    if key != "name":
                        new_policy_config["constraints"][subProfileName][key] = constraint[key]
        new_config[policy_config["policyName"]] = new_policy_config

    log.debug("---------------NEW--------------")
    log.debug(json.dumps(new_config, indent=2))
    log.debug("---------------NEW--------------")

    # Get all policies from vCenter, the objects are VMOMI objects
    old_configs = policy_module.find(
        policy_name=None, service_instance=service_instance, profile=profile
    )

    # make JSON representation of current policies
    # old_configs holds only the rules that are in the scope of interest (provided in argument config_input)
    old_config = {}
    for policy in old_configs:
        if policy["policyName"] not in new_config.keys():
            continue

        policy_json = {"constraints": {}}

        for constraint in policy["constraints"]:
            subProfileName = constraint["name"]
            policy_json["constraints"][subProfileName] = {}
            for key in constraint.keys():
                if key != "name":
                    policy_json["constraints"][subProfileName][key] = constraint[key]

        old_config[policy["policyName"]] = policy_json

    log.debug("--------------OLD---------------")
    log.debug(json.dumps(old_config, indent=2))
    log.debug("--------------OLD---------------")

    # Find rules changes
    changes = []
    rule_diff = drift.drift_report(old_config, new_config, diff_level=0)
    policies_diff = json.loads(json.dumps(rule_diff))  # clone object
    if policies_diff is not None:
        ret["changes"] = policies_diff

        log.debug("==============DRIFT===============")
        log.debug(json.dumps(rule_diff, indent=2))
        log.debug("==============DRIFT===============")

        # add changes for process if not dry-run
        for policy_name in policies_diff:
            new_policy = policies_diff[policy_name]["new"]
            changes.append({**{"name": policy_name}, **new_policy})

    # If it's not dry-run and has changes, then apply changes
    if not __opts__["test"] and changes:
        success = True

        log.debug("===============CHANGES==============")
        log.debug(json.dumps(changes, indent=2))
        log.debug("===============CHANGES==============")

        comments = {}
        for change in changes:
            try:
                # convert change report to configuration object
                new_constraints = []
                for constr_name in change["constraints"].keys():
                    constr_props = change["constraints"][constr_name].copy()
                    constr_props["name"] = constr_name
                    new_constraints.append(constr_props)
                update = {"policyName": change["name"], "constraints": new_constraints}
                # save policy
                policy_module.save(update, service_instance, profile)
                comments[change["name"]] = {
                    "status": "SUCCESS",
                    "message": f"Storage policy '{change['name']}' has been changed successfully.",
                }
            except Exception as err:
                success = False
                comments[change["name"]] = {
                    "status": "FAILURE",
                    "message": f"Error occured while save storage policy '{change['name']}': {err}",
                }

        ret["comment"] = comments  # it's more readable if passed as object
        ret["result"] = success  # at least one success

    return ret



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import salt.exceptions
import saltext.vmware.utils.cluster as utils_cluster
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datacenter as utils_datacenter

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_cluster_drs"
__proxyenabled__ = ["vmware_cluster_drs"]
__func_alias__ = {"get_": "get"}


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def configure(
    cluster,
    datacenter,
    enable=False,
    enable_vm_behavior_overrides=True,
    default_vm_behavior=None,
    vmotion_rate=3,
    advanced_settings=None,
    service_instance=None,
    profile=None,
):
    """
    Configure a Distributed Resource Scheduler (DRS) for a given cluster

    Supported proxies: esxcluster

    cluster
        The cluster name

    datacenter
        The datacenter name to which the cluster belongs

    enable
        Enable DRS for the cluster

    enable_vm_behavior_overrides
        Flag that dictates whether DRS Behavior overrides for individual virtual machines are enabled.
        The default value is true.
        When this flag is true, overrides the default_vm_behavior.
        When this flag is false, the default_vm_behavior value applies to all virtual machines.

    default_vm_behavior
        Specifies the cluster-wide default DRS behavior for virtual machines.
        Valid Values:

        - ``fullyAutomated``: Specifies that VirtualCenter should automate both the migration of virtual machines
          and their placement with a host at power on.

        - ``manual``: Specifies that VirtualCenter should generate recommendations for virtual machine migration
          and for placement with a host, but should not implement the recommendations automatically.

        - ``partiallyAutomated``: Specifies that VirtualCenter should generate recommendations for virtual
          machine migration and for placement with a host, but should automatically
          implement only the placement at power on.

    vmotion_rate
        Threshold for generated ClusterRecommendations. DRS generates only those recommendations that are above
        the specified vmotionRate. Ratings vary from 1 to 5. This setting applies to manual, partiallyAutomated,
        and fullyAutomated DRS clusters. 1 - Conservative, 5 - Aggressive. Default is 3.

    advanced_settings
        Advanced options for the cluster, to be passed in as a dictionary.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_cluster_drs.configure cluster1 dc1 enable=True
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        dc_ref = utils_datacenter.get_datacenter(service_instance, datacenter)
        cluster_ref = utils_cluster.get_cluster(dc_ref=dc_ref, cluster=cluster)
        cluster_spec = vim.cluster.ConfigSpecEx()
        cluster_spec.drsConfig = vim.cluster.DrsConfigInfo()
        cluster_spec.drsConfig.enabled = enable
        cluster_spec.drsConfig.enableVmBehaviorOverrides = enable_vm_behavior_overrides
        cluster_spec.drsConfig.defaultVmBehavior = default_vm_behavior
        cluster_spec.drsConfig.vmotionRate = 6 - vmotion_rate
        cluster_spec.drsConfig.option = []
        for key in advanced_settings or {}:
            cluster_spec.drsConfig.option.append(
                vim.OptionValue(key=key, value=advanced_settings[key])
            )
        utils_cluster.update_cluster(cluster_ref=cluster_ref, cluster_spec=cluster_spec)
    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareRuntimeError) as exc:
        return {cluster: False, "reason": str(exc)}
    return {cluster: True}


def get(cluster_name, datacenter_name, service_instance=None, profile=None):
    """
    Get DRS info about a cluster in a datacenter

    cluster_name
        The cluster name

    datacenter_name
        The datacenter name to which the cluster belongs

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_cluster_drs.get cluster_name=cl1 datacenter_name=dc1
    """
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        dc_ref = utils_datacenter.get_datacenter(service_instance, datacenter_name)
        cluster_ref = utils_cluster.get_cluster(dc_ref=dc_ref, cluster=cluster_name)
        ret["enabled"] = cluster_ref.configurationEx.drsConfig.enabled
        ret[
            "enable_vm_behavior_overrides"
        ] = cluster_ref.configurationEx.drsConfig.enableVmBehaviorOverrides
        ret["default_vm_behavior"] = cluster_ref.configurationEx.drsConfig.defaultVmBehavior
        ret["vmotion_rate"] = 6 - cluster_ref.configurationEx.drsConfig.vmotionRate
        ret["advanced_settings"] = {}
        for obj in cluster_ref.configurationEx.drsConfig.option:
            ret["advanced_settings"][obj.key] = obj.value
    except (salt.exceptions.VMwareApiError, salt.exceptions.VMwareRuntimeError) as exc:
        return {cluster_name: False, "reason": str(exc)}
    return ret


def vm_affinity_rule(
    name,
    affinity,
    vm_names,
    cluster_name,
    datacenter_name,
    enabled=True,
    mandatory=None,
    service_instance=None,
    profile=None,
):
    """
    Configure a virtual machine to virtual machine DRS rule

    name
        The name of the rule.

    affinity
        (boolean) Describes whether to make affinity or anti affinity rule.

    vm_names
        List of virtual machines associated with DRS rule.

    cluster_name
        The name of the cluster to configure a rule on.

    datacenter_name
        The name of the datacenter where the cluster exists.

    enabled
        (optional, boolean) Enable the DRS rule being created. Defaults to True.

    mandatory
        (optional, boolean) Sets whether the rule being created is mandatory. Defaults to False.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_cluster_drs.vm_affinity_rule name="Example Anti-Affinity Rule" affinity=False vm_names='["vm1", "vm2"]' cluster_name=cl1 datacenter_name=dc1 mandatory=True
    """
    log.debug(f"Configuring a vm to vm DRS rule {name} on cluster {cluster_name}.")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_common.get_datacenter(service_instance, datacenter_name)
    cluster_ref = utils_cluster.get_cluster(dc_ref, cluster_name)
    vm_refs = []
    missing_vms = []
    for vm_name in vm_names:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)
        if not vm_ref:
            missing_vms.append(vm_name)
        vm_refs.append(vm_ref)
    if missing_vms:
        raise salt.exceptions.VMwareApiError({f"Could not find virtual machines {missing_vms}"})
    rules = cluster_ref.configuration.rule
    rule_ref = None
    if rules:
        for rule in rules:
            if rule.name == name:
                rule_info = utils_cluster.drs_rule_info(rule)
                if utils_cluster.check_affinity(rule) != affinity:
                    return {
                        "updated": False,
                        "message": f"Existing rule of name {name} has an affinity of {not affinity} and cannot be changed, make new rule.",
                    }
                if (
                    rule_info["vms"] == vm_names
                    and rule_info["enabled"] == enabled
                    and rule_info["mandatory"] == mandatory
                ):
                    return {
                        "updated": True,
                        "message": "Exact rule already exists.",
                    }
                rule_ref = rule

    if rule_ref:
        utils_cluster.update_drs_rule(rule_ref, vm_refs, enabled, mandatory, cluster_ref)
        return {"updated": True}
    else:
        utils_cluster.create_drs_rule(name, affinity, vm_refs, enabled, mandatory, cluster_ref)
        return {"created": True}


def rule_info(cluster_name, datacenter_name, rule_name=None, service_instance=None, profile=None):
    """
    Return a list of all the DRS rules on a given cluster, or one DRS rule if filtered by rule_name.

    cluster_name
        The name of the cluster to get rules from.

    datacenter_name
        The name of the datacenter where the cluster exists.

    rule_name
        (optional) Return only the rule with rule_name

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_cluster_drs.rule_info cluster_name=cl1 datacenter_name=dc1
    """
    log.debug(f"Getting rules info on cluster {cluster_name}.")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    dc_ref = utils_common.get_datacenter(service_instance, datacenter_name)
    cluster_ref = utils_cluster.get_cluster(dc_ref, cluster_name)
    rules = cluster_ref.configuration.rule
    info = []
    if rule_name:
        for rule in rules:
            if rule.name == rule_name:
                return utils_cluster.drs_rule_info(rule)
    else:
        for rule in rules:
            info.append(utils_cluster.drs_rule_info(rule))
        return info
    raise salt.exceptions.VMwareApiError({f"Rule name {rule_name} not found."})



"""
Salt execution module for VPN statistics
Provides methods to Display VPN Statistics and Sessions.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_vpn_statistics"


def __virtual__():
    return __virtualname__


def get_ipsec_statistics(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    locale_service_id,
    service_id,
    session_id,
    tier0_id=None,
    tier1_id=None,
    verify_ssl=True,
    cert=None,
    enforcement_point_path=None,
):
    """
    Retrieves VPN IPSec Statistics from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vpn_statistics.get_ipsec_statistics hostname=nsxt-manager.local  ...


    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which ipsec statistics should be retrieved

    locale_service_id
        id of locale service for example default

    service_id
        id of service for example default

    session_id
        id of session

    Enter one of the tier0 or tier1 id

    tier0_id
        id of tier0 for example vmc

    tier1_id is currently not supported
        id of tier1

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    enforcement_point_path
        (Optional) String Path of the enforcement point

    """
    log.info("Retrieving ipsec statistics for SDDC %s", sddc_id)

    if tier0_id and tier1_id:
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ONE}
    elif not (tier0_id or tier1_id):
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE}

    if tier0_id:
        tier = "tier-0s"
        tier_id = tier0_id
    else:
        tier = "tier-1s"
        tier_id = tier1_id

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{tier}/{tier_id}/locale-services/{locale_service_id}/"
        "ipsec-vpn-services/{service_id}/sessions/{session_id}/statistics"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        tier=tier,
        tier_id=tier_id,
        locale_service_id=locale_service_id,
        service_id=service_id,
        session_id=session_id,
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["enforcement_point_path"], enforcement_point_path=enforcement_point_path
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_vpn_statistics.get_ipsec_statistics",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_ipsec_sessions(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    locale_service_id,
    service_id,
    tier0_id=None,
    tier1_id=None,
    verify_ssl=True,
    cert=None,
    cursor=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Retrieves ipsec session from Given SDDC
        this also include l2vpn sessions

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vpn_statistics.get_ipsec_sessions hostname=nsxt-manager.local  ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which public ips should be retrieved

    locale_service_id
        id of locale service for example default

    service_id
        id of service for example default

    Enter one of the tier0 or tier1 id

    tier0_id
        id of tier0 for example vmc

    tier1_id is currently not supported
        id of tier1

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    """

    log.info("Retrieving ipsec sessions for SDDC %s", sddc_id)

    if tier0_id and tier1_id:
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ONE}
    elif not (tier0_id or tier1_id):
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE}

    if tier0_id:
        tier = "tier-0s"
        tier_id = tier0_id
    else:
        tier = "tier-1s"
        tier_id = tier1_id

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{tier}/{tier_id}/locale-services/{locale_service_id}/"
        "ipsec-vpn-services/{service_id}/sessions"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        tier=tier,
        tier_id=tier_id,
        locale_service_id=locale_service_id,
        service_id=service_id,
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_vpn_statistics.get_ipsec_sessions",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_l2vpn_statistics(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    locale_service_id,
    service_id,
    session_id,
    tier0_id=None,
    tier1_id=None,
    verify_ssl=True,
    cert=None,
    enforcement_point_path=None,
    source=None,
):
    """
    Retrieves L2VPN Statistics from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vpn_statistics.get_l2vpn_statistics hostname=nsxt-manager.local  ...


    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which ipsec statistics should be retrieved

    locale_service_id
        id of locale service for example default

    service_id
        id of service for example default

    session_id
        id of session

    Enter one of the tier0 or tier1 id

    tier0_id
        id of tier0 for example vmc

    tier1_id is currently not supported
        id of tier1

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    enforcement_point_path
        (Optional) String Path of the enforcement point

    source
        (Optional) valid options are realtime, cached
    """

    log.info("Retrieving l2vpn statistics for SDDC %s", sddc_id)

    if tier0_id and tier1_id:
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ONE}
    elif not (tier0_id or tier1_id):
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE}

    if tier0_id:
        tier = "tier-0s"
        tier_id = tier0_id
    else:
        tier = "tier-1s"
        tier_id = tier1_id

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{tier}/{tier_id}/locale-services/{locale_service_id}/"
        "l2vpn-services/{service_id}/sessions/{session_id}/statistics"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        tier=tier,
        tier_id=tier_id,
        locale_service_id=locale_service_id,
        service_id=service_id,
        session_id=session_id,
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["enforcement_point_path", "source"],
        enforcement_point_path=enforcement_point_path,
        source=source,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_vpn_statistics.get_l2vpn_statistics",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_l2vpn_sessions(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    locale_service_id,
    service_id,
    tier0_id=None,
    tier1_id=None,
    verify_ssl=True,
    cert=None,
    cursor=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Retrieves l2vpn session from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vpn_statistics.get_l2vpn_sessions hostname=nsxt-manager.local  ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which public ips should be retrieved

    locale_service_id
        id of locale service for example default

    service_id
        id of service for example default

    Enter one of the tier0 or tier1 id

    tier0_id
        id of tier0 for example vmc

    tier1_id is currently not supported
        id of tier1

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    """

    log.info("Retrieving l2vpn sessions for SDDC %s", sddc_id)

    if tier0_id and tier1_id:
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ONE}
    elif not (tier0_id or tier1_id):
        log.error(vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE)
        return {"error": vmc_constants.VPN_ERROR_SPECIFY_ATLEAST_ONE}

    if tier0_id:
        tier = "tier-0s"
        tier_id = tier0_id
    else:
        tier = "tier-1s"
        tier_id = tier1_id

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{tier}/{tier_id}/locale-services/{locale_service_id}/"
        "l2vpn-services/{service_id}/sessions"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        tier=tier,
        tier_id=tier_id,
        locale_service_id=locale_service_id,
        service_id=service_id,
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_vpn_statistics.get_l2vpn_sessions",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )
